import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import yaml

from paque.task import Section, Task

logger = logging.getLogger("paque.Task")


class Parser(ABC):
    """If you want a different parser, suit yourself. The expected API is a
dictionary of Dict[str, Task], with keys being the names of the tasks"""

    def __init__(self, filename: str):
        self._filename = filename

    @abstractmethod
    def _get_run(self, task_def) -> Optional[str]:
        pass

    @abstractmethod
    def _get_message(self, task_def) -> Optional[str]:
        pass

    @abstractmethod
    def _get_sleep(self, task_def) -> Optional[str]:
        pass

    @abstractmethod
    def _get_depends(self, task_def) -> Optional[List[Task]]:
        pass

    @abstractmethod
    def parse(self) -> Dict[str, Task]:
        pass


class YAMLParser(Parser):
    """Specific parser for YAML files"""

    def __init__(self, filename: str):
        self.filename = filename

    @staticmethod
    def _find_section(dic, section: str) -> Section:
        """Only allows one section with the same name (that's the 0 at the end).
        Casting specific sections to specific types is done explicitly in _build_tasks

        """
        values = list(filter(lambda x: section in x.keys(), dic))
        if len(values) == 0:
            return None
        return values[0][section]

    def _get_run(self, task_def) -> Optional[str]:
        _run = self._find_section(task_def, "run")
        if _run is None:
            return None
        if isinstance(_run, List):
            if all([isinstance(run_item, str) for run_item in _run]):
                return "\n".join(_run)
        if isinstance(_run, str):
            return _run
        raise Exception("Run section should only contain a string or list of strings")

    def _get_sleep(self, task_def) -> Optional[str]:
        """Sleep if cast to an integer _after_ all variables have been evaluated. So,
this can only fail when running (or dry-running )the plan"""
        _sleep = self._find_section(task_def, "sleep")
        if _sleep is None:
            return None
        if isinstance(_sleep, str):
            return _sleep
        if isinstance(_sleep, int):
            return str(_sleep)
        raise Exception(
            "Sleep section should only contain integers or strings (to be interpolated by arguments)"
        )

    def _get_depends(self, task_def) -> Optional[List[Task]]:
        _depends = self._find_section(task_def, "depends")
        if _depends is None:
            return None
        if isinstance(_depends, List):
            if all([isinstance(dependency, str) for dependency in _depends]):
                return [Task(name=name) for name in _depends]
        raise Exception(
            "Depends section should only contain an array of strings (if only one dependency, break it as an array)"
        )

    def _get_message(self, task_def) -> Optional[str]:
        _message = self._find_section(task_def, "message")
        if _message is None:
            return None
        if isinstance(_message, List):
            if all([isinstance(message_item, str) for message_item in _message]):
                return "\n".join(_message)
        if isinstance(_message, str):
            return _message
        raise Exception(
            "Message section should only contain a string or list of strings"
        )

    def _build_tasks(self, parsed_yaml: Dict[str, Any]) -> Dict[str, Task]:
        task_dict = {}
        for task_name, task_def in parsed_yaml.items():
            run: Optional[str] = self._get_run(task_def)
            sleep: Optional[str] = self._get_sleep(task_def)
            message: Optional[str] = self._get_message(task_def)
            depends: Optional[List[Task]] = self._get_depends(task_def)
            task = Task(task_name, run, depends, message, sleep)
            task_dict[task_name] = task
        return task_dict

    def parse(self) -> Dict[str, Task]:
        data = open(self.filename, "r")
        try:
            loaded = yaml.safe_load(data)
        except Exception as e:
            raise Exception("Could not load the YAML file: %s", e)
        return self._build_tasks(loaded)



import logging
from typing import Dict, List, Optional, Union

Section = Optional[Union[List[str], str, int]]

logger = logging.getLogger("paque.Task")


class Task:
    """Task instance. Having it as a separate entity (instead of faking it with
dictionaries inside of the Planner) means it's easier to test, debug and swap
implementations. Obviously"""

    def __init__(
        self,
        name: str = "",
        run: Optional[str] = None,
        depends: Optional[List["Task"]] = None,
        message: Optional[str] = None,
        sleep: Optional[str] = None,
    ):
        self.name = name
        self.run = run
        self.depends = depends
        self.message = message
        self.sleep = sleep

    def __repr__(self) -> str:
        """Why did you use emoji? Why not?"""
        depends_then_runs = f"[{self.depends}] -> {self.run}"
        says_sleeps = f"(🗣 {self.message},😴 {self.sleep})"
        return f"Task({self.name}: {depends_then_runs} {says_sleeps}"

    def __lt__(self, other):
        if not isinstance(other, Task):
            return NotImplemented
        return repr(self) < repr(other)

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Task):
            return NotImplemented
        return self.name == other.name and str(self) == str(other)

    def replace_dependencies_with(self, steps: List["Task"]) -> "Task":
        """Replace dependencies with fully processed dependencies. This is to ensure
the final plan has all tasks as the final state requires. This is actually an
advantage when faking it with dictionaries inside the Planner: you can do this
by just replacements as you go. Not with a separate Task implementation with
properties.

        """
        proper_dependencies = []
        if self.depends is None:
            return self
        for dependency in self.depends:
            done_step = [step for step in steps if step.name == dependency.name][0]
            proper_dependencies.append(done_step)
        self.depends = proper_dependencies
        return self

    @staticmethod
    def _args_dict(args: List[str]) -> Dict[str, str]:
        """Helper to convert arguments to dictionries for formatting/replacement"""
        args_dict = {}
        for arg in args:
            key, value = arg.split(":")
            args_dict[key] = value
        return args_dict

    def with_args(self, args: Optional[List[str]]) -> "Task":
        """Replaces the properties of this task with the replacements after argument
substitution"""

        def interpolate(item):
            try:
                return item.format(**args_dict)
            except KeyError as e:
                raise Exception(f"Argument not found: {e}")

        logger.debug("Task: %s", self)
        logger.debug("args: %s", args)

        if args is None:
            return self

        args_string = " ".join(args)

        args_dict: Dict[str, str] = self._args_dict(args)
        self.name = interpolate(self.name)
        if args_string not in self.name:
            self.name = self.name + " {}".format(args_string)
        if self.run is not None:
            self.run = interpolate(self.run)
        if self.message is not None:
            self.message = interpolate(self.message)
        if self.sleep is not None:
            self.sleep = interpolate(self.sleep)
        if self.depends is not None:  # Depends can be an empty list instead of None!
            self.depends = [
                dependency.with_args(args)
                for dependency in self.depends
                if dependency is not None
            ]

        logger.debug("interpolated Task: %s", self)
        return self

    def get_sleep(self) -> Optional[int]:
        """Sleep is special: it eventually needs to be an integer. We cast it at the
very end of the processes"""
        if self.sleep is None:
            return None
        try:
            return int(self.sleep)
        except ValueError as v:
            raise Exception("Sleep should be an integer by this point %s", v)



import logging
import subprocess
import time
from typing import Any, List

logger = logging.getLogger("paque.executor")


class Executor:
    def __init__(self, plan: List[Any]) -> None:
        self._plan = plan

    def dry_run(self) -> None:
        logger.debug("The plan: %s", self._plan)
        for task in self._plan:
            logger.info(">>> Running task %s", task.name)
            msgs = zip(
                ["message", "run", "sleep"], [task.message, task.run, task.get_sleep()]
            )
            for msg, log in msgs:
                logger.info("%s: %s", msg, log)

    def run(self) -> None:
        logger.info("Running plan")
        for task in self._plan:
            message = task.message
            if message is not None:
                logger.info(message)
            run = task.run
            if run is not None:
                logger.debug("Running %s", run)
                subprocess.run(run, shell=True, check=True)
            duration = task.get_sleep()
            if duration is not None:
                logger.debug("Sleeping for %s", duration)
                time.sleep(duration)



import logging
from typing import Any, Dict, List, Optional, Tuple, Union

from paque.task import Task

logger = logging.getLogger("paque.planner")


class Planner:
    def __init__(self, _tasks: Dict[str, Task]) -> None:
        self._tasks = _tasks
        logger.debug(self._tasks)
        self._steps: List[Task] = []

    def find_dependencies(self, task: str) -> List[str]:
        if task in self._tasks.keys():
            dependencies = self._tasks[task].depends
            logger.debug("DEPENDENCIES: %s", dependencies)
            if isinstance(dependencies, List):
                return [d.name for d in dependencies if d is not None]
            if isinstance(dependencies, Task):
                return [dependencies.name]
            if dependencies is None:
                return []
        raise Exception("Task required, and not found", task)

    @staticmethod
    def dependency_and_arguments(
        raw_dependency: str,
    ) -> Union[Tuple[str, List[str]], Tuple[str, None]]:
        if " " in raw_dependency:
            dependency, *args = raw_dependency.split(" ")
            return dependency, args
        return raw_dependency, None

    def _with_args(self, task_name: str, args: Optional[List[str]]) -> Task:
        logger.debug("Task to replace: %s", task_name)
        extracted = self._tasks[task_name]
        if args is not None:
            extracted.with_args(args)
        return extracted

    def _remaining(self, dependencies: List[str]) -> List[str]:
        done = {task.name for task in self._steps}
        return sorted(set(dependencies).difference(done))

    def _plan(self, task_name: str, args: Optional[List[str]] = None) -> None:
        initial_state = len(self._steps)
        logger.debug("Already ran: %s", self._steps)
        task = self._with_args(task_name, args)
        initial_dependencies = self.find_dependencies(task_name)
        dependencies = self._remaining(initial_dependencies)
        if len(dependencies) == 0:
            logger.debug("Can run %s(%s) with no dependencies left", task, args)
            step = self._with_args(task_name, args)
            if step not in self._steps:
                logger.debug("Adding step: %s", step)
                self._steps.append(self._with_args(task_name, args))
            return

        logger.debug("Has dependencies, plan them: %s", dependencies)
        for raw_dependency in dependencies:
            dependency, new_args = self.dependency_and_arguments(raw_dependency)
            logger.debug("Dependency %s has args %s", dependency, new_args)
            self._plan(dependency, new_args)

        if len(self._steps) > initial_state and len(self._remaining(dependencies)) != 0:
            logger.debug("Process has progressed, recurse on same task")
            self._plan(task_name, args)
        if len(self._remaining(dependencies)) == 0:
            logger.debug("We have parsed dependencies, fix and repeat on same task")
            self._plan(task_name, args)
        if len(self._steps) == initial_state:
            logger.debug("We have stopped improving")
            raise Exception("Could not find a plan to solve {}({})".format(task, args))
        logger.debug("Final recursion")
        self._plan(task_name, args)

    def plan(self, task: str) -> List[Task]:
        logger.info(">>> Planning execution for task %s", task)
        self._plan(task)
        abbreviated_plan = [task.name for task in self._steps]
        logger.info(">>> Plan requires %s", abbreviated_plan)
        correct_steps = []
        for step in self._steps:
            correct_step = step.replace_dependencies_with(self._steps)
            correct_steps.append(correct_step)
        return correct_steps



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['paque']

package_data = \
{'': ['*']}

install_requires = \
['cleo>=0.8.0,<0.9.0', 'colorlog>=4.1.0,<5.0.0', 'pyyaml>=5.3,<6.0']

entry_points = \
{'console_scripts': ['paque = paque.paque:main']}

setup_kwargs = {
    'name': 'paque',
    'version': '0.1.0',
    'description': 'paque simplifies running simple workflows you want to run. It offers a few features of `make`, but removing most of its power',
    'long_description': '# Paque `¯\\_(ツ)_/¯`\n\n<!-- markdown-toc start - Don\'t edit this section. Run M-x markdown-toc-refresh-toc -->\n\n- [Paque `¯\\_(ツ)_/¯`](#paque-ツ)\n    - [Installation](#installation)\n        - [Tab completion](#tab-completion)\n    - [Why?](#why)\n    - [What?](#what)\n    - [How?](#how)\n    - [FAQ](#faq)\n        - [Why YAML and not FOOML?](#why-yaml-and-not-fooml)\n        - [Is this production ready?](#is-this-production-ready)\n        - [What\'s with the name?](#whats-with-the-name)\n    - [Future development](#future-development)\n\n<!-- markdown-toc end -->\n\nPaque simplifies running simple workflows you want to run. It offers a few\nfeatures of `make`, but removing most of its power. It runs on a `paquefile` or\n`paquefile.yaml` (or just pass the name of the file). You can see a simple\n[example in the root folder](paquefile.yaml).\n\nIt supports Python 3.6.5+ (for no particular reason aside from being my default,\nit should work just fine on any relatively recent Python 3)\n\n## Installation\n\nIt should be enough to run\n\n```bash\npip install paque\n```\n\nand then \n\n```bash\npaque taskname  # If you have a paquefile or paquefile.yaml file\n```\n\n### Tab completion\n\nThis is a reminder to myself to implement tab completion (at lest for ZSH, since\nI have done that before)\n\n## Why?\n\nI had a series of `Dockerfile` I wanted to build sequentially, and run some\ncommands. It wasn\'t a right fit for docker-compose, so I wrote a makefile\n([here](https://github.com/rberenguel/spark_hadoop_kudu/blob/master/makefile)).\nThe result was excellent, BUT if you want to add logging, or any kind of\ninformation you need to start resorting to "makefile hacks" (like passing\nvariables down the dependency stack, or accessing subpaths of requirements...)\nthat didn\'t feel right.\n\nSo, I decided to write this.\n\n## What?\n\nYou define your tasks in a YAML file with specific syntax, of the form:\n\n```yaml\ntaskname:\n  - run: "what it runs"\n  - message: "what it logs"\n  - depends:\n      - task_it_depends_1\n      - task_it_depends_2\n  - sleep: integer\n\notherwise:\n  - run:\n      - "You can run several commands"\n      - "Passing them as an array"\n  - message:\n      - "Likewise for logging"\n      - "Yes."\n```\n\nYou can also use arguments and multiple arguments,\n\n```yaml\ntaskname:\n  - run: "{something} {folder}"\n  - message: "This does {something} on {folder}"\n\nmain:\n  - depends:\n      - taskname folder:/Users/foo/ something:rmdir\n```\n\nFor now you can\'t have spaces in arguments. Sorry.\n\nFor usage, you would just \n\n```bash\npaque taskname\n```\n\n## How?\n\nYAML (following the rules above) is converted into a dictionary of task names\nand [Tasks](paque/task.py) by a [Parser](paque/parser.py). Then a simple\ndepth-first-search [planner](paque/planner.py) finds an execution that satisfies\nall dependencies and transitive dependencies (with arguments) of `taskname`.\nFinally, the plan is passed to an [executor](paque/executor.py) that offloads it\nto the shell (or just logs it).\n\n## FAQ\n\n### Why YAML and not FOOML?\n\nI find YAML pretty readable and writable, as long as you restrict what you can\ndo. Since there is no nesting here, you can\'t shoot yourself in the foot with\nYAML. If you really can\'t stand YAML, you have two options\n\n- Use [dhall](https://github.com/dhall-lang/dhall-lang) and convert from dhall to YAML (_recommended_)\n- Write a [parser](paque/parser.py) for your favourite markup\n\n### Is this production ready?\n\nWell… `¯\\_(ツ)_/¯` I\'m pretty sure there is an issue with argument replacement\nin a corner case, but I can\'t put my finger on which. For simple use cases, this\nshould be safe. Since there is no branching, there is not much that can go wrong\nthough.\n\n### What\'s with the name?\n\nFor one, it\'s `python+make=pake`, but it was taken (I should have checked\nbefore), so left it at `paque`. It\'s also a form of "pa\' qué", an Spanish slang\nfor "para qué". I.e. _what for?_ `¯\\_(ツ)_/¯`\n\n### Contributing\n\nI\'m happy to receive PRs, so don\'t be shy. Also let me know if you used it, that\ncould be fun. As you see from the root folder, you will need to use\n[poetry](https://github.com/python-poetry/poetry)\n\n## Future development\n\nI will keep using it, so any bugs I find will be fixed. Likewise, I will keep\nimproving it, although the current version is "almost enough". Currently on the\n"roadmap" I have:\n\n- Generate a graphviz plot of the plan(s). This was one of the motivations to\n  write my own thingy, after all\n- Better tests: I wrote the ones I have with a combination of TDD and "let\'s\n  test and print". I want less tests of happy paths and more tests of corner\n  cases\n- Custom exceptions, right now it\'s just "raise that"\n- Automatically convert paquefile.dhall into a YAML paquefile (this was supposed\n  to be in this version but I got lazy)\n- Have a nicer CLI (probably using [cleo](https://github.com/sdispater/cleo))\n- Possibly, running tasks in parallel (this is a hard one given how the planner\n  works, so probably won\'t)\n- Conditionals?\n- Fixing the bug that is likely there in argument substitution\n',
    'author': 'Ruben Berenguel',
    'author_email': 'ruben+poetry@mostlymaths.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rberenguel/paque',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import argparse
import logging
import os

from colorlog import ColoredFormatter  # type: ignore

from paque.executor import Executor
from paque.parser import YAMLParser
from paque.planner import Planner

logger = logging.getLogger("paque")


def configure_logger():
    formatter = ColoredFormatter(
        "%(log_color)s%(levelname)s - %(message)s",
        datefmt=None,
        reset=True,
        log_colors={
            "DEBUG": "yellow",
            "INFO": "cyan",
            "WARNING": "purple",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
        secondary_log_colors={},
        style="%",
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def get_paquefile(args):
    paquefile = args.paquefile

    def check_file(candidate: str):
        try:
            candidate_path = os.path.join(os.getcwd(), candidate)
            with open(candidate_path) as _:
                logger.debug("File found: %s", candidate_path)
                return True
        except IOError:
            logger.debug("File not found: %s (IO failed)", candidate_path)
            return False
        logger.debug("File not found: %s (no exception?)", candidate_path)
        return False

    if paquefile is None:

        candidates = ["paquefile", "paquefile.yaml"]
        try:
            indexer = list(map(check_file, candidates)).index(True)
        except ValueError:
            indexer = None
        if indexer is None:
            raise Exception(
                "No file provided and neither paquefile, paquefile.yaml are available"
            )
        return candidates[indexer]
    if check_file(paquefile):
        return paquefile


def paque(args):
    if args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    paquefile = get_paquefile(args)
    parser = YAMLParser(paquefile)
    planner = Planner(parser.parse())
    if args.dry_run:
        Executor(planner.plan(args.task[0])).dry_run()
    else:
        Executor(planner.plan(args.task[0])).run()


def main(**args):
    configure_logger()
    parser = argparse.ArgumentParser(description="Paque: Not make, but in Python")
    parser.add_argument(
        "paquefile",
        metavar="paquefile",
        type=str,
        help="File to run",
        nargs="?",
        default=None,
    )
    parser.add_argument(
        "task",
        metavar="task",
        type=str,
        help="Task to run from file",
        nargs=1,
        default=None,
    )
    parser.add_argument(
        "--dry-run", dest="dry_run", action="store_true", help="Dry run the plan"
    )
    parser.add_argument(
        "--debug", dest="debug", action="store_true", help="Set logging to debug level"
    )
    args = parser.parse_args()
    try:
        paque(args)
    except Exception as exc:
        logger.error(exc)


if __name__ == "__main__":
    main()
